// Base64 functions

#include "base64.h"

#include <stdlib.h>

static const char base64_table[] = {
	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
	'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b',
	'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p',
	'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3',
	'4', '5', '6', '7', '8', '9', '+', '/'
};

char *base64_encode_buffer(const char *s, size_t size) {
	size_t encoded_size = (size + 2) / 3 * 4 + 1;
	char *encoded = malloc(encoded_size);
	if (!encoded)
		return NULL;
	char *e = encoded;
	for (;;) {
		unsigned char c = (*s >> 2) & 0x3f;
		*e++ = base64_table[c];
		c = (*s++ << 4) & 0x30;
		if (!--size) {
			*e++ = base64_table[c];
			*e++ = '=';
			*e++ = '=';
			break;
		}
		c |= (*s >> 4) & 0x0f;
		*e++ = base64_table[c];
		c = (*s++ << 2) & 0x3c;
		if (!--size) {
			*e++ = base64_table[c];
			*e++ = '=';
			break;
		}
		c |= (*s >> 6) & 0x03;
		*e++ = base64_table[c];
		c = *s++ & 0x3f;
		*e++ = base64_table[c];
		if (!--size)
			break;
	}
	*e = '\0';
	return encoded;
}

unsigned char base64_value(char c) {
	if ('A' <= c && c <= 'Z')
		return c - 'A';
	if ('a' <= c && c <= 'z')
		return c - 'a' + 26;
	if ('0' <= c && c <= '9')
		return c - '0' + 52;
	switch (c) {
	case '+':
		return 62;
	case '/':
		return 63;
	}
	return 0xff;
}

bool is_base64(char c) {
	return c == '+' || ('/' <= c && c <= '9') || ('A' <= c && c <= 'Z')
		|| ('a' <= c && c <= 'z');
}
