// UTF-8 handling

#include "utf8.h"

static void utf8_length_action(unsigned code, void *data);
static void utf8_valid_action(unsigned code, void *data);

size_t utf8_encode(unsigned code, char *encoded) {
	char *e = encoded;
	if (code < 0x00000080)
		*e++ = code;
	else {
		if (code < 0x00000800)
			*e++ = 0xc0 | code >> 6;
		else {
			if (code < 0x00010000) {
				if (0x0000d800 <= code && code <= 0x0000dfff)
					return 0;
				*e++ = 0xe0 | code >> 12;
			}
			else if (code < 0x00110000) {
				*e++ = 0xf0 | code >> 18;
				*e++ = 0x80 | (code >> 12 & 0x3f);
			}
			else
				return 0;
			*e++ = 0x80 | (code >> 6 & 0x3f);
		}
		*e++ = 0x80 | (code & 0x3f);
	}
	return e - encoded;
}

size_t utf8_length(const char *s, _Bool *error) {
	size_t length = 0;
	*error = utf8_parse(s, utf8_length_action, &length);
	return length;
}

static void utf8_length_action(__attribute__((unused)) unsigned code,
	void *data) {
	size_t *length = data;
	(*length)++;
}

_Bool utf8_parse(const char *s, utf8_action action, void *data) {
	while (*s) {
		unsigned code;
		if (*s & 0x80) {
			if (!(*s & 0x40))
				return 1;
			int class = 1;
			if (*s & 0x20) {
				class++;
				if (*s & 0x10) {
					class++;
					if (*s & 0x08)
						return 1;
				}
			}
			code = *s++ & (0x3f >> class);
			int i;
			for (i = 0; i < class; i++) {
				if ((*s & 0xc0) != 0x80)
					return 1;
				code <<= 6;
				code |= *s++ & 0x3f;
			}
			switch (class) {
			case 1:
				if (code < 0x00000080)
					return 1;
				break;
			case 2:
				if (code < 0x00000800)
					return 1;
				if (0x0000d800 <= code && code <= 0x0000dfff)
					return 1;
				break;
			case 3:
				if (code < 0x00010000)
					return 1;
				if (code >= 0x00110000)
					return 1;
				break;
			}
		}
		else
			code = *s++;
		action(code, data);
	}
	return 0;
}

_Bool utf8_valid(const char *s) {
	return !utf8_parse(s, utf8_valid_action, NULL);
}

static void utf8_valid_action(__attribute__((unused)) unsigned code,
	__attribute__((unused)) void *data) {
}
